//This header file contains all of the functions that are used for
//the implementation of geometric algebra. These are described 
//fully in my report. It also contains the functions that are required
//for the ray tracer, i.e defining a plane from a polygon, line 
//plane intersection, and finding a point in a triangle.
//Furthermore, I have included functions that allow for the addition
//and subtraction of multivectors, their division by a scalar, and
//their normalisation. 


#ifndef __FUNCTIONS_H__
#define __FUNCTIONS_H__

#include <iostream>
#include <cmath>
#include "polygon.h"


mv reverse(mv);

mv inverse(mv);

mv dual(mv);
mv meet(mv, mv);
mv projection(mv, mv);
mv rejection(mv, mv);

mv add(mv, mv);
mv sub(mv, mv);
mv sub(vector, mv);
vector sub(vector, vector);
mv scal_div(mv, double);
mv normalise(mv);

vector line_plane_intersect(line, plane);
vector line_reflection_in_plane(line, mv);

mv line_line_intersect(line, line);

plane find_plane(polygon);

bool inside_edge(vector, vector, vector, vector);
bool inside_tri(vector, polygon);

mv reverse(mv a)
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	result.setScalar(a1);
	result.sete1(a2);
	result.sete2(a3);
	result.sete3(a4);
	result.sete12(-a5);
	result.sete23(-a6);
	result.sete31(-a7);
	result.sete123(-a8);

	return result;

}


mv inverse(mv a)
{
	mv store_1, store_2, result;
	double denom;

	denom = (a.getScalar()*a.getScalar()) + (a.gete1()*a.gete1()) + (a.gete2()*a.gete2()) + (a.gete3()*a.gete3()) + (a.gete12()*a.gete12()) + (a.gete23()*a.gete23()) + (a.gete31()*a.gete31()) + (a.gete123()*a.gete123());

	if(denom == 0)		//special case for the division by 0
	{
		return a;
	}

	else
	{
		store_1 = reverse(a);
	
		result.setScalar(store_1.getScalar()/denom);
		result.sete1(store_1.gete1()/denom);
		result.sete2(store_1.gete2()/denom);
		result.sete3(store_1.gete3()/denom);
		result.sete12(store_1.gete12()/denom);
		result.sete23(store_1.gete23()/denom);
		result.sete31(store_1.gete31()/denom);
		result.sete123(store_1.gete123()/denom);

		return result;
	}
}


mv dual(mv a)
{
	mv result;

	mv iPseudo(0,0,0,0,0,0,0,1);
	
	iPseudo = inverse(iPseudo);

	result = gp(a, iPseudo);

	return result;
}


mv meet(mv a, mv b)		//returns the meet between the two supplied multivectors
{
	mv result;

	mv Da = dual(a);

	result = ip_cont(Da,b);

	return result;
}


mv projection(mv a, mv b)	//returns the projection of a onto b
{
	mv result = gp(ip_hest(a,b), inverse(b));
	
	return result;
}


mv rejection(mv a, mv b)	//returns the rejection of a on b
{
	mv result = gp(op(a,b), inverse(b));
	
	return result;
}


mv add(mv a, mv b)		//addition of two multivectors
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8  = b.gete123();

	result.setScalar(a1+b1);

	result.sete1(a2+b2);

	result.sete2(a3+b3);

	result.sete3(a4+b4);

	result.sete12(a5+b5);

	result.sete23(a6+b6);

	result.sete31(a7+b7);

	result.sete123(a8+b8);

	return result;
}

mv sub(mv a, mv b)		//subtraction of two multivectors
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8  = b.gete123();

	result.setScalar(a1-b1);

	result.sete1(a2-b2);

	result.sete2(a3-b3);

	result.sete3(a4-b4);

	result.sete12(a5-b5);

	result.sete23(a6-b6);

	result.sete31(a7-b7);

	result.sete123(a8-b8);

	return result;
}



mv sub(vector vec_a, mv b)		//subtraction of a multivector from a vector
{
	mv result;

	double a_store[3];

	a_store[0] = vec_a.gete1();	
	a_store[1] = vec_a.gete2();
	a_store[2] = vec_a.gete3();

	mv a(0, a_store[0], a_store[1], a_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	result.setScalar(a1-b1);

	result.sete1(a2-b2);

	result.sete2(a3-b3);

	result.sete3(a4-b4);

	result.sete12(a5-b5);

	result.sete23(a6-b6);

	result.sete31(a7-b7);

	result.sete123(a8-b8);

	return result;
}


mv sub(mv a, vector vec_b)		//subtraction of a vector from a multivector
{
	mv result;

	double b_store[3];

	b_store[0] = vec_b.gete1();	
	b_store[1] = vec_b.gete2();
	b_store[2] = vec_b.gete3();

	mv b(0, b_store[0], b_store[1], b_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	result.setScalar(a1-b1);

	result.sete1(a2-b2);

	result.sete2(a3-b3);

	result.sete3(a4-b4);

	result.sete12(a5-b5);

	result.sete23(a6-b6);

	result.sete31(a7-b7);

	result.sete123(a8-b8);

	return result;
}


vector sub(vector vec_a, vector vec_b)		//subtraction of two vectors
{	
	vector result;

	result.sete1(vec_a.gete1() - vec_b.gete1());

	result.sete2(vec_a.gete2() - vec_b.gete2());

	result.sete3(vec_a.gete3() - vec_b.gete3());

	return result;
}



mv scal_div(mv a, double b)			//scalar division of a multivector
{	
	mv result;

	double a1 = a.getScalar()/b;
	double a2 = a.gete1()/b;
	double a3 = a.gete2()/b;
	double a4 = a.gete3()/b;
	double a5 = a.gete12()/b;
	double a6 = a.gete23()/b;
	double a7 = a.gete31()/b;
	double a8 = a.gete123()/b;

	result.setScalar(a1);

	result.sete1(a2);

	result.sete2(a3);

	result.sete3(a4);

	result.sete12(a5);

	result.sete23(a6);

	result.sete31(a7);

	result.sete123(a8);

	return result;
}

	
mv normalise(mv a)			//normailisation of a vector. 
{
	mv result;			

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();
	
	double denom = sqrt(pow(a1,2) + pow(a2,2) + pow(a3,2) + pow(a4,2) + pow(a5,2) + pow(a6,2) + pow(a7,2) + pow(a8,2));

	if(denom==0)			//special case if the denominator is equal to 0
	return a;

	a1 = a1/denom;
	a2 = a2/denom;
	a3 = a3/denom;
	a4 = a4/denom;
	a5 = a5/denom;
	a6 = a6/denom;
	a7 = a7/denom;
	a8 = a8/denom;

	result.setall(a1,a2,a3,a4,a5,a6,a7,a8);

	return result;
}


vector line_plane_intersect(line l, plane pl)		//this function returns the point of intersection between
{							//a pline and a plane in the form of a position vector.
	vector l_point, l_direction, result;
	mv numerator, denominator, p_intersect;


	l_point = l.getp();
	l_direction = l.getDirection();

	numerator = op(l_point, pl.getbivector());
	numerator = dual(numerator);
	numerator.setScalar(numerator.getScalar() - pl.getoffset());
	numerator = gp(numerator, l_direction);

	denominator = op(l_direction, pl.getbivector());
	denominator = dual(denominator);
	denominator = inverse(denominator);	

	p_intersect = gp(numerator, denominator);

	p_intersect = sub(l_point, p_intersect);

	result.sete1(p_intersect.gete1());
	result.sete2(p_intersect.gete2());
	result.sete3(p_intersect.gete3());

	return result;
}



vector line_reflection_in_plane(line l, mv plane)	//this function was not described in my report
{							//but will return a vector that is the reflection
	vector l_point, l_direction, result;		//of the vector given as a parameter in the lane
	mv numerator, denominator, reflect_dir;		//provided.

	l_point = l.getp();
	l_direction = l.getDirection();

	reflect_dir = gp(plane,l_direction);
	reflect_dir = gp(reflect_dir, plane);
	
	result.sete1(reflect_dir.gete1());
	result.sete2(reflect_dir.gete2());
	result.sete3(reflect_dir.gete3());

	return result;
}


mv line_line_intersect(line a, line b)				//this function was not described in my report
{								//but will return the point of intersetion
	mv left, left_inv, right, right_inv, result;		//of two lines that lie on the same plane.

	left_inv = op(a.getDirection(), b.getDirection());
	left_inv = inverse(left_inv);

	left = op(b.getp(), b.getDirection());
	left = gp(left, left_inv);
	left = gp(left, a.getDirection());

	right_inv = op(b.getDirection(), a.getDirection());
	right_inv = inverse(right_inv);

	right = op(a.getp(), a.getDirection());
	right = gp(right, right_inv);
	right = gp(right, b.getDirection());

	result = add(left, right);

	return result; 
}


bool inside_edge(vector p, vector v1, vector v2, vector v3)	//this function is called by inside_tri. its
{								//parameters are three vectors describing a 
	vector edge_1 = sub(v1, v2);				//triangle and a fourth vector defining a point. 
	vector edge_2 = sub(v1, v3);				//It will return true if the point is on the 
	vector edge_p = sub(v1, p);				//on the correct side of the line v1v2 to be inside
								//the triangle.
	mv bi_p = op(edge_1, edge_p);	
	mv bi_tri = op(edge_1, edge_2);

	mv inner = ip_cont(bi_p, bi_tri);

	if(inner.getScalar() <= 0)
	return true;

	else
	return false;
}


bool inside_tri(vector p, polygon tri)		//This function accepts a point and a polygon as its parameters.
{						//It makes use of inside_edge to determine whether the point is
	vector v1 = tri.getv1();		//inside the polygon.
	vector v2 = tri.getv2();
	vector v3 = tri.getv3();

	if(inside_edge(p, v1, v2, v3) && inside_edge(p, v3, v1, v2) && inside_edge(p, v2, v3, v1))
	return true;
	else
	return false;
}


plane find_plane(polygon p)		//this final function accepts a polygon as its arguement
{					//and will return the plane on which the polygon lies.
	plane result;

	vector v1 = p.getv1();
	vector v2 = p.getv2();
	vector v3 = p.getv3();
	
	
	double offset;

	mv move(0,v1.gete1(),v1.gete2(),v1.gete3(),0,0,0,0);

	v2 = sub(v2,v1);
	v3 = sub(v3,v1);
	v1 = sub(v1,v1);

	mv bi = op(v2,v3);

	mv normal = dual(bi);

	mv project = rejection(move, bi);

	offset = sqrt( pow(project.gete1(),2) + pow(project.gete2(),2) + pow(project.gete3(),2) );
	
	bi = normalise(bi);

	result.setall(-offset, bi);

	return result;
}

#endif