//This header file forms the basis of the geometric algebra calculations.
//It hold all of the functions that define the fundamental products
//includeing the inner, outer, and geometric products.
//Each function has been overloaded several times so as to allow for their
//use between any combination of multivectors and vectors.
//To see how each has been calculated, please refer to the calculation
//tables in my report.


#ifndef __OPERATORS_H__
#define __OPERATORS_H__

#include <iostream>
#include <cmath>
#include "polygon.h"

using namespace std;

mv gp(mv, mv);			//geometric products
mv gp(vector, vector);
mv gp(mv, vector);
mv gp(vector, mv);

mv op(mv, mv);			//outer products
mv op(vector, vector);
mv op(mv, vector);
mv op(vector, mv);

mv ip_hest(mv, mv);		//hestenes (default) inner product
mv ip_cont(mv, mv);		//left contraction inner product


mv gp(mv a, mv b)
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	
	result.setScalar( a1 * b1 + a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7 - a8 * b8 );

	result.sete1( a2 * b1 + a1 * b2 + a5 * b3 - a7 * b4 - a3 * b5 - a8 * b6 + a4 * b7 - a6 * b8 );
	
	result.sete2( a3 * b1 - a5 * b2 + a1 * b3 + a6 * b4 + a2 * b5 - a4 * b6 - a8 * b7 - a7 * b8 );

	result.sete3( a4 * b1 + a7 * b2 - a6 * b3 + a1 * b4 - a8 * b5 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a5 * b1 - a3 * b2 + a2 * b3 + a8 * b4 + a1 * b5 + a7 * b6 - a6 * b7 + a4 * b8 );

	result.sete23( a6 * b1 + a8 * b2 - a4 * b3 + a3 * b4 - a7 * b5 + a1 * b6 + a5 * b7 + a2 * b8 );
	
	result.sete31( a7 * b1 + a4 * b2 + a8 * b3 - a2 * b4 + a6 * b5 - a5 * b6 + a1 * b7 + a3 * b8 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8 );

	
	return result;
}


mv gp(vector vec_a, vector vec_b)
{
	mv result;

	double a_store[3];
	double b_store[3];

	a_store[0] = vec_a.gete1();	
	a_store[1] = vec_a.gete2();
	a_store[2] = vec_a.gete3();

	b_store[0] = vec_b.gete1();	
	b_store[1] = vec_b.gete2();
	b_store[2] = vec_b.gete3();

	mv a(0, a_store[0], a_store[1], a_store[2], 0, 0, 0, 0);
	mv b(0, b_store[0], b_store[1], b_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	
	result.setScalar( a1 * b1 + a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7 - a8 * b8 );

	result.sete1( a2 * b1 + a1 * b2 + a5 * b3 - a7 * b4 - a3 * b5 - a8 * b6 + a4 * b7 - a6 * b8 );
	
	result.sete2( a3 * b1 - a5 * b2 + a1 * b3 + a6 * b4 + a2 * b5 - a4 * b6 - a8 * b7 - a7 * b8 );

	result.sete3( a4 * b1 + a7 * b2 - a6 * b3 + a1 * b4 - a8 * b5 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a5 * b1 - a3 * b2 + a2 * b3 + a8 * b4 + a1 * b5 + a7 * b6 - a6 * b7 + a4 * b8 );

	result.sete23( a6 * b1 + a8 * b2 - a4 * b3 + a3 * b4 - a7 * b5 + a1 * b6 + a5 * b7 + a2 * b8 );
	
	result.sete31( a7 * b1 + a4 * b2 + a8 * b3 - a2 * b4 + a6 * b5 - a5 * b6 + a1 * b7 + a3 * b8 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8 );

	return result;
}


mv gp(mv a, vector vec_b)
{
	mv result;

	double b_store[3];

	b_store[0] = vec_b.gete1();	
	b_store[1] = vec_b.gete2();
	b_store[2] = vec_b.gete3();

	mv b(0, b_store[0], b_store[1], b_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	
	result.setScalar( a1 * b1 + a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7 - a8 * b8 );

	result.sete1( a2 * b1 + a1 * b2 + a5 * b3 - a7 * b4 - a3 * b5 - a8 * b6 + a4 * b7 - a6 * b8 );
	
	result.sete2( a3 * b1 - a5 * b2 + a1 * b3 + a6 * b4 + a2 * b5 - a4 * b6 - a8 * b7 - a7 * b8 );

	result.sete3( a4 * b1 + a7 * b2 - a6 * b3 + a1 * b4 - a8 * b5 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a5 * b1 - a3 * b2 + a2 * b3 + a8 * b4 + a1 * b5 + a7 * b6 - a6 * b7 + a4 * b8 );

	result.sete23( a6 * b1 + a8 * b2 - a4 * b3 + a3 * b4 - a7 * b5 + a1 * b6 + a5 * b7 + a2 * b8 );
	
	result.sete31( a7 * b1 + a4 * b2 + a8 * b3 - a2 * b4 + a6 * b5 - a5 * b6 + a1 * b7 + a3 * b8 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8 );

	return result;
}


mv gp(vector vec_a, mv b)
{
	mv result;

	double a_store[3];

	a_store[0] = vec_a.gete1();	
	a_store[1] = vec_a.gete2();
	a_store[2] = vec_a.gete3();

	mv a(0, a_store[0], a_store[1], a_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	
	result.setScalar( a1 * b1 + a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7  - a8 * b8 );

	result.sete1( a2 * b1 + a1 * b2 + a5 * b3 - a7 * b4 - a3 * b5 - a8 * b6 + a4 * b7 - a6 * b8 );
	
	result.sete2( a3 * b1 - a5 * b2 + a1 * b3 + a6 * b4 + a2 * b5 - a4 * b6 - a8 * b7 - a7 * b8 );

	result.sete3( a4 * b1 + a7 * b2 - a6 * b3 + a1 * b4 - a8 * b5 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a5 * b1 - a3 * b2 + a2 * b3 + a8 * b4 + a1 * b5 + a7 * b6 - a6 * b7 + a4 * b8 );

	result.sete23( a6 * b1 + a8 * b2 - a4 * b3 + a3 * b4 - a7 * b5 + a1 * b6 + a5 * b7 + a2 * b8 );
	
	result.sete31( a7 * b1 + a4 * b2 + a8 * b3 - a2 * b4 + a6 * b5 - a5 * b6 + a1 * b7 + a3 * b8 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8 );

	return result;
}


mv ip_cont(mv a, mv b)
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8  = b.gete123();

	result.setScalar( a1 * b1 + a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7  - a8 * b8 );

	result.sete1( a1 * b2 - a3 * b5 + a4 * b7 - a6 * b8 );
	
	result.sete2( a1 * b3 + a2 * b5 - a4 * b6 - a7 * b8 );

	result.sete3( a1 * b4 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a1 * b5 + a4 * b8 );

	result.sete23( a1 * b6 + a2 * b8 );
	
	result.sete31( a1 * b7 + a3 * b8 );

	result.sete123( a1 * b8 );

	return result;

}



mv ip_hest(mv a, mv b)
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8  = b.gete123();

	result.setScalar( a2 * b2 + a3 * b3 + a4 * b4 - a5 * b5 - a6 * b6 - a7 * b7  - a8 * b8 );

	result.sete1( a5 * b3 - a7 * b4 - a3 * b5 - a8 * b6 + a4 * b7 - a6 * b8 );
	
	result.sete2( - a5 * b2 + a6 * b4 + a2 * b5 - a4 * b6 - a8 * b7 - a7 * b8 );

	result.sete3( a7 * b2 - a6 * b3 - a8 * b5 + a3 * b6 - a2 * b7 - a5 * b8 );

	result.sete12( a8 * b4 + a4 * b8 );

	result.sete23( a8 * b2 + a2 * b8 );
	
	result.sete31( a8 * b3 + a3 * b8);

	result.sete123( 0 );	

	return result;

}



mv op(mv a, mv b)
{
	mv result;

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8  = b.gete123();

	result.setScalar(a1 * b1 );

	result.sete1( a1 * b2 + a2 * b1 );

	result.sete2( a1 * b3 + a3 * b1 );

	result.sete3( a1 * b4 + a4 * b1);

	result.sete12( a1 * b5 + a2 * b3 - a3 * b2 + a5 * b1 );

	result.sete23( a1 * b6 + a3 * b4 - a4 * b3 + a6 * b1 );

	result.sete31( a1 * b7 - a2 * b4  + a4 * b2 + a7 * b1 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8  );

	return result;

}


mv op(vector vec_a, vector vec_b)
{
	mv result;

	double a_store[3];
	double b_store[3];

	a_store[0] = vec_a.gete1();	
	a_store[1] = vec_a.gete2();
	a_store[2] = vec_a.gete3();

	b_store[0] = vec_b.gete1();	
	b_store[1] = vec_b.gete2();
	b_store[2] = vec_b.gete3();

	mv a(0, a_store[0], a_store[1], a_store[2], 0, 0, 0, 0);
	mv b(0, b_store[0], b_store[1], b_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	result.setScalar(a1 * b1 );

	result.sete1( a1 * b2 + a2 * b1 );

	result.sete2( a1 * b3 + a3 * b1 );

	result.sete3( a1 * b4 + a4 * b1);

	result.sete12( a1 * b5 + a2 * b3 - a3 * b2 + a5 * b1 );

	result.sete23( a1 * b6 + a3 * b4 - a4 * b3 + a6 * b1 );

	result.sete31( a1 * b7 - a2 * b4  + a4 * b2 + a7 * b1 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8  );

	return result;
}


mv op(mv a, vector vec_b)
{
	mv result;

	double b_store[3];

	b_store[0] = vec_b.gete1();	
	b_store[1] = vec_b.gete2();
	b_store[2] = vec_b.gete3();

	mv b(0, b_store[0], b_store[1], b_store[2], 0, 0, 0, 0);

	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	result.setScalar(a1 * b1 );

	result.sete1( a1 * b2 + a2 * b1 );

	result.sete2( a1 * b3 + a3 * b1 );

	result.sete3( a1 * b4 + a4 * b1);

	result.sete12( a1 * b5 + a2 * b3 - a3 * b2 + a5 * b1 );

	result.sete23( a1 * b6 + a3 * b4 - a4 * b3 + a6 * b1 );

	result.sete31( a1 * b7 - a2 * b4  + a4 * b2 + a7 * b1 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8  );

	return result;
}


mv op(vector vec_a, mv b)
{
	mv result;

	double a_store[3];

	a_store[0] = vec_a.gete1();	
	a_store[1] = vec_a.gete2();
	a_store[2] = vec_a.gete3();

	mv a(0, a_store[0], a_store[1], a_store[2], 0, 0, 0, 0);


	double a1 = a.getScalar();
	double a2 = a.gete1();
	double a3 = a.gete2();
	double a4 = a.gete3();
	double a5 = a.gete12();
	double a6 = a.gete23();
	double a7 = a.gete31();
	double a8 = a.gete123();

	double b1 = b.getScalar();
	double b2 = b.gete1();
	double b3 = b.gete2();
	double b4 = b.gete3();
	double b5 = b.gete12();
	double b6 = b.gete23();
	double b7 = b.gete31();
	double b8 = b.gete123();

	result.setScalar(a1 * b1 );

	result.sete1( a1 * b2 + a2 * b1 );

	result.sete2( a1 * b3 + a3 * b1 );

	result.sete3( a1 * b4 + a4 * b1);

	result.sete12( a1 * b5 + a2 * b3 - a3 * b2 + a5 * b1 );

	result.sete23( a1 * b6 + a3 * b4 - a4 * b3 + a6 * b1 );

	result.sete31( a1 * b7 - a2 * b4  + a4 * b2 + a7 * b1 );

	result.sete123( a8 * b1 + a6 * b2 + a7 * b3 + a5 * b4 + a4 * b5 + a2 * b6 + a3 * b7 + a1 * b8  );

	return result;

}

#endif
