//This library defines functions that allow for some arbitrary 
//transformations of vectors. Although they are not used by the
//raytracer I have implemented them as they could be used in
//an extension of this project to create a lighting model.


#ifndef __TRANSFORMATIONS_H__
#define __TRANSFORMATIONS_H__

#include <cmath>

using namespace std;

vector reflect(vector, vector);
vector rotate(vector, vector, vector);
vector normalize(vector);

vector reflect(vector ref, vector ref2)		//this function will return the reflection of ref1 in
{						//ref2.
	double result[3];

	vector ref_in = normalize(ref2);

	mv store = gp(ref_in, ref);
	mv reflected = gp(store, ref_in);

	result[0] = reflected.gete1();	
	result[1] = reflected.gete2();	
	result[2] = reflected.gete3();	

	vector out(result[0], result[1], result[2]); 

	return out;
}


vector rotate(vector rot, vector ref1, vector ref2)	//this function uses two reflections to obtain 
{							//a rotation of vector rot about a point. 
	double result[3];				//It is first reflected in ref1. The result of
	vector store;					//this is then reflected in ref2 to obtain
	vector out;					//a rotation.
							//This is a demonstration of how geometric 
	store = reflect(rot, ref1);			//algebra instinctively incorporates quoternions.
	out = reflect(store, ref2);

	result[0] = out.gete1();
	result[1] = out.gete2();
	result[2] = out.gete3();

	out.setall(result[0], result[1], result[2]);

	return out;
}


vector normalize(vector in)			//a simple function to normalise a vector.
{
	double store[3];
	double magnitude;
	vector out;

	store[0] = in.gete1();
	store[1] = in.gete2();
	store[2] = in.gete3();

	magnitude = sqrt((store[0]*store[0]) + (store[1]*store[1]) + (store[2]*store[2]));

	store[0] /= magnitude;
	store[1] /= magnitude;
	store[2] /= magnitude; 

	out.setall(store[0], store[1], store[2]);

	return out;
}

#endif

