#include <iostream>
#include <GL/glut.h>

#include "gl_funcs.h"
#include "multivector.h"
#include "vector.h"
#include "line.h"
#include "plane.h"
#include "operators.h"
#include "functions.h"
#include "transformations.h"
#include "polygon.h"

using namespace std;

int ga();

vector vertex[5000];		//array for storing all the vectors in the obj file.
polygon polys[3000];		//array for storing all the polys in the obj file.
plane planes[3000];		//array for storing the planes on which the polys lie
line rays[1000000];		//array for storing each ray that intersects the mesh
vector intersections[1000000];	//array for storing the points of intersection with the mesh

vector ortho_cam_pos(0,0,0);	//defines the position of the camera

string obj_file;		//string to store the name of the obj file
string image_file;		//string to store the name of the output file
string x_resolution;		
string y_resolution;
string image_size;		//this string is used to define the size of the image using 
				//Macgick++.h

int x_res;	
int y_res;

int intersect_count = 0;	
int poly_count = 0;

double upper_limit = 0;		//stores the e2 position of the highest vertex
double lower_limit = 0;		//stores the e2 position of the lowest vertex
double left_limit = 0;		//stores the e1 position of the left most vertex
double right_limit = 0;		//stores the e1 position of the right most vertex
double back_limit = 0;		//stores the e3 position of the furthest vertex
double front_limit = 0;		//stores the e3 position of the nearest vertex

#include "obj_import.h"
#include "ortho_raytrace.h"

vector d(0,0.5,1);		//used for drawing the viewing angles of the camera
vector e(0,-0.5,1);
vector f(0.5,0,1);
vector g(-0.5,0,1);

line A(ortho_cam_pos,d);
line B(ortho_cam_pos,e);
line C(ortho_cam_pos,f);
line D(ortho_cam_pos,g);



int ga()
{	
	open_obj();		//import .obj

	raytrace();		//perform raytrace

	return 1;	
}


void display()
{
	int i;

	glClearColor(0.7, 0.7, 0.7, 0);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

	glTranslatef(0,0,-zoom);
	glTranslatef(tx,ty,0);
	glRotatef(rotx,1,0,0);
	glRotatef(roty,0,1,0);	


	glColor3f(1.0, 1.0, 1.0);	// draw horizontal grid
	glBegin(GL_LINES);
	for(i=-10;i<=10;++i) 
	{
		glVertex3f(i,0,-10);
		glVertex3f(i,0,10);

		glVertex3f(10,0,i);
		glVertex3f(-10,0,i);
	}
	glEnd();

	/*glColor3f(1.0, 1.0, 1.0);	// draw veritcal grid
	glBegin(GL_LINES);
	for(i=-10;i<=10;++i) 
	{
		glVertex3f(-10,i,0);
		glVertex3f(10,i,0);

		glVertex3f(i,10,0);
		glVertex3f(i,-10,0);
	}
	glEnd();*/

	obj_draw();			//draw object
	
	glColor3f(1.0, 0.0, 0.0);
	for(i=0; i<intersect_count; i++)	//draw the rays that intersect the object
	{
		rays[i].draw_intersect_from_origin(intersections[i]);
	}


	glColor3f(0.0, 0.0, 0.0);
	A.draw_from_origin();
	B.draw_from_origin();
	C.draw_from_origin();
	D.draw_from_origin();

	glutSolidSphere(0.2,10,10);

	glutSwapBuffers();
}






int main(int argc, char* argv[])
{
	if(argc != 5)		//if there are the wrong number of arguements
	{
		cout<<"Wrong number of parameters"<<endl;
		exit(0);	//quit
	}

	obj_file = argv[1];	//place the command line arguements in the appropriate strings	
	obj_file = "./obj/" + obj_file;
	image_file = argv[2];
	image_file = "./images/" + image_file;
	x_resolution = argv[3];
	y_resolution = argv[4];

	x_res = atoi(argv[3]);	//set the resolution
	y_res = atoi(argv[4]);

	image_size = x_resolution + "x" + y_resolution;		//the string image_size has to be in the 
								//form x_resolution x y_resolution. 
								//E.g. 100x100
								//This is the form that the image magick 
								//library accepts as its resolution


	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
	glutInitWindowSize(1000,1000);
	glutInitWindowPosition(100,100);
	glutCreateWindow("Geometric Algebra");


	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);

	Init();

	ga();

	glutMainLoop();
}
